#include <stdio.h>
#include "glext.h"
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include "common.h"
#include "geometry.h"

GLuint cover_texture;

extern unsigned int g_iPhotonIDMap, g_iPhotonDirMap, g_iPhotonPowerMap;
extern PhotonMap *g_pPhotonMap;

CGcontext context;
CGprogram photonMapVert, photonMapFrag;
CGprogram simpleVert, simpleFrag;
CGprofile vertexProfile, fragmentProfile;

static void handleCgError() 
{
	char message[128];
	sprintf(message, "Cg error: %s\n", cgGetErrorString(cgGetError()));
	reportError(message);
	exit(1);
}

void initCg()
{
	// set up the context
	cgSetErrorCallback(handleCgError);
	context = cgCreateContext();
	//cgGLRegisterStates(context);
	cgGLSetManageTextureParameters(context, CG_TRUE);

	// set up the profiles
	if (cgGLIsProfileSupported(CG_PROFILE_VP40))
		vertexProfile = CG_PROFILE_VP40;
	else 
	{
		//reportError("vp40 vertex profile is not supported on this system.\n");
		vertexProfile = CG_PROFILE_ARBVP1;
	}

	if (cgGLIsProfileSupported(CG_PROFILE_FP40))
		fragmentProfile = CG_PROFILE_FP40;
	else
	{
		//reportError("fp40 vertex profile is not supported on this system.\n");
		fragmentProfile = CG_PROFILE_ARBFP1;
	}

	// load the programs
    photonMapVert = cgCreateProgramFromFile(context, CG_SOURCE, "Shaders/photonMapVert.cg",
                                            vertexProfile, NULL, NULL);
    if (!cgIsProgramCompiled(photonMapVert))
        cgCompileProgram(photonMapVert);

    // Enable the appropriate vertex profile and load the vertex program.
    cgGLEnableProfile(vertexProfile);
    cgGLLoadProgram(photonMapVert);

    // And similarly set things up for the fragment program.
    photonMapFrag = cgCreateProgramFromFile(context, CG_SOURCE, "Shaders/photonMapFrag.cg",
                                              fragmentProfile, NULL, NULL);
    if (!cgIsProgramCompiled(photonMapFrag))
        cgCompileProgram(photonMapFrag);

    cgGLEnableProfile(fragmentProfile);
    cgGLLoadProgram(photonMapFrag);

	simpleVert = cgCreateProgramFromFile(context, CG_SOURCE, "Shaders/simpleVert.cg",
                                            vertexProfile, NULL, NULL);
    if (!cgIsProgramCompiled(simpleVert))
        cgCompileProgram(simpleVert);
    cgGLEnableProfile(vertexProfile);
    cgGLLoadProgram(simpleVert);

	simpleFrag = cgCreateProgramFromFile(context, CG_SOURCE, "Shaders/simpleFrag.cg",
                                            fragmentProfile, NULL, NULL);
    if (!cgIsProgramCompiled(simpleFrag))
        cgCompileProgram(simpleFrag);
    cgGLEnableProfile(fragmentProfile);
    cgGLLoadProgram(simpleFrag);

}

void setUpTextures()
{
	glBindTexture(GL_TEXTURE_2D, g_iPhotonIDMap);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 
		g_pPhotonMap->getPhotonIDTextureWidth(), 
		g_pPhotonMap->getPhotonIDTextureHeight(),
		0, GL_RGBA, GL_UNSIGNED_BYTE,
		g_pPhotonMap->getIDTexture());

	glBindTexture(GL_TEXTURE_2D, g_iPhotonDirMap);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 
		g_pPhotonMap->getTextureWidth(), 
		g_pPhotonMap->getTextureHeight(),
		0, GL_RGBA, GL_UNSIGNED_BYTE,
		g_pPhotonMap->getDirTexture());

	glBindTexture(GL_TEXTURE_2D, g_iPhotonPowerMap);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 
		g_pPhotonMap->getTextureWidth(), 
		g_pPhotonMap->getTextureHeight(),
		0, GL_RGBA, GL_UNSIGNED_BYTE,
		g_pPhotonMap->getPowerTexture());
}

void drawObjectCg(Mesh *pMesh)
{
	setUpTextures();
    cgGLBindProgram(photonMapVert);
    cgGLBindProgram(photonMapFrag);

    // Bind uniform parameters to vertex shader
    cgGLSetStateMatrixParameter(cgGetNamedParameter(photonMapVert, "ModelViewProj"),
                                CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                CG_GL_MATRIX_IDENTITY);
    cgGLSetStateMatrixParameter(cgGetNamedParameter(photonMapVert, "ModelView"),
                                CG_GL_MODELVIEW_MATRIX,
                                CG_GL_MATRIX_IDENTITY);
    cgGLSetStateMatrixParameter(cgGetNamedParameter(photonMapVert, "ModelViewIT"),
                                CG_GL_MODELVIEW_MATRIX,
                                CG_GL_MATRIX_INVERSE_TRANSPOSE);

	CGparameter param = cgGetNamedParameter(photonMapVert, "input.position");
	cgGLEnableClientState(param);
	cgGLSetParameterPointer(param, 3, GL_FLOAT, 0, pMesh->m_pVertex);
	param = cgGetNamedParameter(photonMapVert, "input.normal");
	cgGLEnableClientState(param);
	cgGLSetParameterPointer(param, 3, GL_FLOAT, 0, pMesh->m_pNormal);
	param = cgGetNamedParameter(photonMapVert, "input.texUV");
	cgGLEnableClientState(param);
	cgGLSetParameterPointer(param, 2, GL_FLOAT, 0, pMesh->m_pTexCoord);
	param = cgGetNamedParameter(photonMapVert, "input.photonID");
	cgGLEnableClientState(param);
	cgGLSetParameterPointer(param, 2, GL_FLOAT, 0, g_pPhotonMap->getTexCoord());

	float photonIDDimensions[4] = { g_pPhotonMap->getPhotonIDTextureWidth(),
		g_pPhotonMap->getPhotonIDTextureHeight(),
		2, 2 };
    cgGLSetParameter4fv(cgGetNamedParameter(photonMapFrag, "photonIDDimensions"), photonIDDimensions);
	float photonDirDimensions[4] = { g_pPhotonMap->getTextureWidth(),
		g_pPhotonMap->getTextureHeight(),
		0, 0 };
    cgGLSetParameter4fv(cgGetNamedParameter(photonMapFrag, "photonDirDimensions"), photonDirDimensions);


	// set up the textures
//    param = cgGetNamedParameter(fragmentProgram, "colorMap");
//	cgGLSetTextureParameter(param, cover_texture);
    param = cgGetNamedParameter(photonMapFrag, "photonIDMap");
	cgGLSetTextureParameter(param, g_iPhotonIDMap);
    param = cgGetNamedParameter(photonMapFrag, "photonDirMap");
	cgGLSetTextureParameter(param, g_iPhotonDirMap);
    param = cgGetNamedParameter(photonMapFrag, "photonPowMap");
	cgGLSetTextureParameter(param, g_iPhotonPowerMap);

//	glActiveTextureARB(GL_TEXTURE0_ARB);
//	glEnable(GL_TEXTURE_2D);
//	glBindTexture(GL_TEXTURE_2D, cover_texture);
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, g_iPhotonIDMap);
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, g_iPhotonDirMap);
	glActiveTextureARB(GL_TEXTURE2_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, g_iPhotonPowerMap);

	glDrawElements(GL_TRIANGLES, pMesh->m_iNumberTriangles * 3, GL_UNSIGNED_INT, pMesh->m_pIndex);
}

void simpleCg()
{
    cgGLBindProgram(simpleVert);
    cgGLBindProgram(simpleFrag);

    cgGLSetStateMatrixParameter(cgGetNamedParameter(simpleVert, "ModelViewProj"),
                                CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                CG_GL_MATRIX_IDENTITY);
}

void drawPhotonDensity()
{
	int i;

	simpleCg();
	glPointSize(3.0f);
	glBegin(GL_POINTS);
	for(i=0; i<g_pPhotonMap->getNumberPhotons(); i++)
	{
		Photon *pPhoton = g_pPhotonMap->getPhoton(i);
		glColor3f(pPhoton->m_fRed, pPhoton->m_fGreen, pPhoton->m_fBlue);
		//glColor3f(0.39f,0,0.39f);
		glVertex3f(pPhoton->m_fX, pPhoton->m_fY, pPhoton->m_fZ);
	}
	glEnd();
}

void drawPhotonTrace()
{
	int i;

	simpleCg();
	//glLineSizef(3.0f);
	glBegin(GL_LINES);
	for(i=0; i<g_pPhotonMap->getNumberPhotons(); i++)
	{
		Photon *pPhoton = g_pPhotonMap->getPhoton(i);
		glColor3f(pPhoton->m_fRed, pPhoton->m_fGreen, pPhoton->m_fBlue);
		//glColor3f(0.39f,0,0.39f);
		glVertex3f(pPhoton->m_fOriginX, pPhoton->m_fOriginY, pPhoton->m_fOriginZ);
		glVertex3f(pPhoton->m_fX, pPhoton->m_fY, pPhoton->m_fZ);
	}
	glEnd();
}