#include "common.h"
#include "meshmanager.h"

extern int g_uiWindowWidth;
extern int g_uiWindowHeight;

extern int g_iNumberPhotons;
extern int g_iNumberLights;
extern int g_iNumberTracePasses;
extern PhotonMap *g_pPhotonMap;
extern Light *g_pLight;
extern float g_fFPS;
extern bool g_bShowHUD;
extern MeshManager *g_pMeshManager;


int time = 0, timebase = 0;
int g_iFrame = 0;

void seedg(int iSeedNumber)
{
	srand(iSeedNumber);
}

// return a float between 0-1
float rand01()
{
	float fNumber = (rand()%RAND_MAX) / float(RAND_MAX);
	return fNumber;
}

void normalize(float x, float y, float z, float &nx, float &ny, float &nz)
{
	float length = sqrt(x*x + y*y + z*z);
	nx = x/length;
	ny = y/length;
	nz = z/length;
}

void reportError(char *string)
{
	printf("ERROR: %s\n", string);
}

// Heron's formula
// Area=SQRT(s(s-a)(s-b)(s-c)),
// where s=(a+b+c)/2 or perimeter/2. 
float calculateArea(Vector3 &v1, Vector3 &v2, Vector3 &v3)
{
	float area;
	float a,b,c,s;

	a = sqrt((v1.x - v2.x)*(v1.x - v2.x) + 
		(v1.y - v2.y)*(v1.y - v2.y) + 
		(v1.z - v2.z)*(v1.z - v2.z));
	b = sqrt((v1.x - v3.x)*(v1.x - v3.x) + 
		(v1.y - v3.y)*(v1.y - v3.y) + 
		(v1.z - v3.z)*(v1.z - v3.z));
	c = sqrt((v2.x - v3.x)*(v2.x - v3.x) + 
		(v2.y - v3.y)*(v2.y - v3.y) + 
		(v2.z - v3.z)*(v2.z - v3.z));

	s = (a + b + c)/2;

	area = sqrt(s * (s - a) * (s - b) * (s - c));

	return area;
}

void setOrthographicProjection() {

	// switch to projection mode
	glMatrixMode(GL_PROJECTION);
	// save previous matrix which contains the 
	//settings for the perspective projection
	glPushMatrix();
	// reset matrix
	glLoadIdentity();
	// set a 2D orthographic projection
	gluOrtho2D(0, g_uiWindowWidth, 0, g_uiWindowHeight);
	// invert the y axis, down is positive
	glScalef(1, -1, 1);
	// mover the origin from the bottom left corner
	// to the upper left corner
	glTranslatef(0, -g_uiWindowHeight, 0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
}

void resetPerspectiveProjection() {
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void renderString(int x, int &y, char *str)
{
	char *c;

	setOrthographicProjection();
	simpleCg();
	glRasterPos2f(x, y);
	for (c=str; *c != '\0'; c++) 
	{
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, *c);
	}
	resetPerspectiveProjection();
	y += 10;
}

void drawHUD()
{
	char string[128];
	int x = 5, y = 30;
	glColor3f(0.9f, 0.8f, 0.2f);
	if(g_bShowHUD)
	{
		renderString(x, y, "HUD (Press H to hide it)");

		sprintf(string, "FPS: %.3f (%.1f ms)", g_fFPS, 1000.0f/g_fFPS);
		renderString(x, y, string);
		renderString(x, y, "---Photon Map---");
		sprintf(string, "Number Lights: %d", g_iNumberLights);
		renderString(x, y, string);
		sprintf(string, "Number Photons: %d", g_iNumberPhotons);
		renderString(x, y, string);
		sprintf(string, "Number Trace Passes: %d", g_iNumberTracePasses);
		renderString(x, y, string);
		sprintf(string, "Texture size - PhotonID: %d %d", g_pPhotonMap->getPhotonIDTextureWidth(),
			g_pPhotonMap->getPhotonIDTextureHeight());
		renderString(x, y, string);
		sprintf(string, "Texture size - Dir&Power: %d %d", g_pPhotonMap->getTextureWidth(),
			g_pPhotonMap->getTextureHeight());
		renderString(x, y, string);
		renderString(x, y, "---Lights---");
		sprintf(string, "Light Position: %.3f %.3f %.3f", g_pLight->m_fX, g_pLight->m_fY, g_pLight->m_fZ );
		renderString(x, y, string);
		sprintf(string, "Light Color: %.3f %.3f %.3f", g_pLight->m_fRed, g_pLight->m_fGreen, g_pLight->m_fBlue );
		renderString(x, y, string);
		renderString(x, y, "---Geometry---");
		int iNumberVertices = 0;
		int iNumberTriangles = 0;
		int i;
		for(i=0; i<g_pMeshManager->getNumberMeshes(); i++)
		{
			iNumberVertices += g_pMeshManager->getMesh(i)->m_iNumberVertices;
			iNumberTriangles += g_pMeshManager->getMesh(i)->m_iNumberTriangles;
		}
		sprintf(string, "Number Meshes: %d", g_pMeshManager->getNumberMeshes());
		renderString(x, y, string);
		sprintf(string, "Number Vertices: %d", iNumberVertices);
		renderString(x, y, string);
		sprintf(string, "Number Triangles: %d", iNumberTriangles);
		renderString(x, y, string);
		renderString(x, y, "---Performance---");
		int iNumTimings = Timing::getNumberTimings();
		for(int i=0; i<iNumTimings; i++)
		{
			Timing kT = Timing::reportTiming(i);
			sprintf(string, "%s: %dms (%d %.2f)", kT.m_sName, kT.m_iDuration, kT.m_iCount, kT.m_iDuration/(float)kT.m_iCount);
			renderString(x, y, string);
		}
	}
	else
	{
		renderString(x, y, "HUD (Press H to show it)");
	}

	Timing::beginFrame();
}

void calculateFPS()
{
	g_iFrame++;
	time=glutGet(GLUT_ELAPSED_TIME);
	
	if (time - timebase > 1000) {
		g_fFPS = g_iFrame*1000.0/(time-timebase);
	 	timebase = time;		
		g_iFrame = 0;
	}
}

Vector3 reflection(Vector3 &kDir, Vector3 &kPos, Vector3 &kNormal)
{
	Vector3 kRef;

	kDir.normalize();
	kNormal.normalize();
	
	kRef.x = 2 * kPos.x * (kNormal.y * kDir.y + kNormal.z * kDir.z) + kDir.x * (kNormal.x * kPos.x - kNormal.y * kPos.y - kNormal.z * kPos.z);
	kRef.y = 2 * kPos.y * (kNormal.x * kDir.x + kNormal.z * kDir.z) + kDir.y * (-kNormal.x * kPos.x + kNormal.y * kPos.y - kNormal.z * kPos.z);
	kRef.z = 2 * kPos.z * (kNormal.x * kDir.x + kNormal.y * kDir.y) + kDir.z * (-kNormal.x * kPos.x - kNormal.y * kPos.y + kNormal.z * kPos.z);

	return kRef;
}